#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <getopt.h>  /* getopt_long() */

#include <fcntl.h>  /* low-level i/o */
#include <unistd.h>
#include <errno.h>
#include <malloc.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/wait.h>

#include <errno.h>

#include <limits.h>

#include <asm/types.h>  /* for videodev2.h */

#include <linux/videodev2.h>

#include <pthread.h>

#include "image_io.h"
#include "tools.h"

#define CLEAR(x) memset(&(x), 0, sizeof(x))

#define IMG 0
#define MIN_HOLD 1
#define PEAK_HOLD 2
#define AVE 3
#define DIF 4
#define HIST 5
#define STRETCH 6
#define TRIG 7
#define DAY_DIR 8
#define MD_LOG_NAME 9
#define MD_LOG_DATA 10
#define STACK_LOG_NAME 11
#define STACK_LOG_DATA 12

#define FNAME_SIZE 160

#define NUM_MD_DIRS 10000

/* Minimum delay between motion detection events */
#define DELAY_BETWEEN_MD 2

struct v4l2_format fmt;

struct buffer {
  void *start;
  size_t length;
};

struct frame {
  unsigned char *img_buffer;
  int width;
  int height;
  time_t timestamp;
  long usec;
};

struct img_stack {
  unsigned long *img_buffer;
  int width;
  int height;
  time_t timestamp;
  long usec;
  int num_imgs;
  unsigned long min_val;
  unsigned long max_val;
};

/* Image frames */

/* Current unpacked frame */
struct frame current_frame;

/* Ring memory for triggered capture */
struct frame *md_frames = NULL;
/* Minimum stack */
struct frame *min_hold; /* Pointer to the stack in use */
struct frame *min_hold_save; /* Pointer to the stack being saved */
struct frame min_hold1;
struct frame min_hold2;
/* Maximum stack */
struct frame *peak_hold; /* Pointer to the stack in use */
struct frame *peak_hold_save; /* Pointer to the stack being saved */
struct frame peak_hold1;
struct frame peak_hold2;
/* Average stack */
struct img_stack *ave_stack; /* Pointer to the stack in use */
struct img_stack *ave_stack_save; /* Pointer to the stack being saved */
struct img_stack ave_stack1;
struct img_stack ave_stack2;

static int img_in_mmap_buffer = 2;

/* Number of images used for each stack in use */
static int num_img_in_stack = 0;
/* Number of images in stacks being saved */
static int num_img_in_save_stack = 0;

/* Length of peak-hold/average/min stack in seconds */
static int stack_length = 60;

static char *dev_name = NULL;
static int fd = -1;

// Skip the first frame of the stack, it tends to be empty
unsigned char first_frame = 1;

/* MMAP buffers */
struct buffer *mmap_buffers = NULL;
static unsigned int n_mmap_buffers = 0;

/* Which type of images will be saved */
static int use_min = 0;
static int use_max = 0;
static int use_ave = 0;
static int use_dif = 0;
static int use_trig = 0;

static int verbose = 0;

/* Thread for saving the ring memory */
pthread_t trig_save_thread;
/* Thread for saving stacks */
pthread_t stack_save_thread;
/* Thread for processing motion detection */
pthread_t md_trig_thread;

/* Motion detection ring memory size */
static int img_in_md_buffer = 25;

pthread_t stack_thread, trig_thread;

/* Index of ring memory slot currently being used */
static int md_i = 0;

/* Threshold used in motion detection */
int md_threshold = 1000;

/* If set, only print the motion detection result, do not save any images */
int md_test = 0;

/* Number of motion regocnition zones in horizontal and vertical directions */
int md_num_x_zones = 12;
int md_num_y_zones = 9;

/* Sums of pixel values for each motion detection zones for previous and new image */
int *md_zone_sums = NULL;
int *md_zone_sums_new = NULL;

/* Number of valid pixels in each motion detection zone */
unsigned int *md_zone_num_pix = NULL;

/* Vector to hold information on which motion detection zone the pixel belongs to */
int *md_zone_reference = NULL;

/* Filename of a mask that can be used to exclude pixels from motion detection */
char *md_mask_fname = NULL;

/* Holds the index where detection happened. If no detection, value is -1. Value > -1 disables motion detection.
 */
int md_detection = -1;

/* Calculate how many images ago the triggering happened */
int md_images_since_detection = -1;

/* Latest triggered event */
time_t md_last_detection = 0;

/* Percentage of images to be saved that were imaged after the triggering event */
int md_percentage = 70;

/* Date of the frame triggering the motion detection
*/
int md_trig_year = 0;
int md_trig_month = 0;
int md_trig_day = 0;

/* Number of motion detection events for the day
 */
int md_trig_number = 0;

/* Zone number that caused the triggering
*/
int md_triggered_zone = -1;

/* Time when the imaging started */
static time_t start_time;

/* Length of this imaging session */
static unsigned long capture_length = ULONG_MAX;

/* Starting time of the stacks */
static time_t stack_timestamp;
static long stack_usec;

/* Ending time of the stacks */
static time_t stack_end_timestamp;
static long stack_end_usec;

/* Output directory path. '/' will be appended automatically */
static char *out_dir = NULL;

/* Prefix, which will be prepended to each filename. '_' will be added. */
static char *img_prefix = NULL;

/* Pid for ring memory saving child proces */
pid_t save_pid;

/* Timestamp for forked process. Set to -1 if no child process exist */
time_t save_timestamp = -1;

/* Keep track if we are saving the motion detection ring memory */
int md_saving_images = 0;

/* Keep track if we are saving image stacks */
int saving_stacks = 0;

/* Keep track, if the md_process() thread is alive */
int md_process_on = 0;

/* Require a full buffer before starting motion detection */
int md_buffer_ready = 0;

/* Log file to save daily data on triggered images */
char *md_log_fname = NULL;
FILE *md_log_fid = NULL;

/* Log file to save data on stacks */
char *stack_log_fname = NULL;
FILE *stack_log_fid = NULL;

/* Default to save both logs */
int save_log = 1;

/* Commandline options */
static const char short_options [] = "c:hn:s:mMadtT:x:y:k:GP:p:o:l:N";

/* Long commandline options */
static const struct option long_options [] = {
  { "capture-device", required_argument, NULL, 'c' }, 
  { "help", no_argument, NULL, 'h' }, 
  { "buffer", required_argument, NULL, 'n' }, 
  { "stack-length", required_argument, NULL, 's' }, 
  { "min", no_argument, NULL, 'm' }, 
  { "max", no_argument, NULL, 'M' }, 
  { "ave", no_argument, NULL, 'a' }, 
  { "dif", no_argument, NULL, 'd' }, 
  { "trigger", no_argument, NULL, 't' },
  { "threshold", required_argument, NULL, 'T' },
  { "x-zones", required_argument, NULL, 'x'},
  { "y-zones", required_argument, NULL, 'y'},
  { "mask", required_argument, NULL, 'k' },
  { "trigger-test", no_argument, NULL, 'G' },
  { "percentage", required_argument, NULL, 'P' },
  { "prefix", required_argument, NULL, 'p' }, 
  { "out-dir", required_argument, NULL, 'o' }, 
  { "capture-length", required_argument, NULL, 'l' }, 
  { "no-log", no_argument, NULL, 'N' }, 
  { "verbose", no_argument, NULL, 'V' }, 
  { 0, 0, 0, 0 }
};

